#ifndef brash_commandline_header_included_p
#define brash_commandline_header_included_p
//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include "BrashException.h"
#include "BrashRedirection.h"
#include <cxxtls/stringexpression.h>
#include <memory>
#include <cxxtls/foreach.h>
#include <map>



struct BrashCommandLine
    ///
    ///  A BrashCommandLine object represents the invocation of non-structural command in the brash language.
    ///  This object, when executed, executes a brash "command line".  Syntactically, command
    ///  line statements look like this:
    ///
    ///@begincode
    ///
    ///   [var1=value1 var2=value2 ...]  CommandNameOrPath   [arguments ...]   [FileRedirections]
    ///
    ///@endcode
    ///
    ///  Structural commands include "if", "while", "for", etc.  The non-structural commands are those that
    ///  represent the executions of seeminingly normal commands:
    ///
    ///   @li  shell builtin functions
    ///   @li  shell functions
    ///   @li  external commands (ie programs and scripts to be executed)
    ///
    ///   The shell builtin commands include:
    ///
    ///   @li   hash
    ///   @li   echo
    ///   @li   set
    ///   @li   env
    ///   @li   jobs
    ///   @li   etc
{
    typedef std::list<cxxtls::StringExpression*> ARGUMENTS_LIST;

    std::auto_ptr<ARGUMENTS_LIST > arguments_;

    // typedef std::list< std::pair<RedirType, cxxtls::StringExpression*> > REDIRECTION_LIST;

    typedef RedirType::REDIRECTION_LIST REDIRECTION_LIST;

    std::auto_ptr< REDIRECTION_LIST > redirections_;

    struct ValueAndIndex
    {
       cxxtls::StringExpression* value_;
       cxxtls::StringExpression* index_;  // could be null, index of array index expression

       ValueAndIndex(cxxtls::StringExpression* value,
                     cxxtls::StringExpression* index=0
                    )
       :  value_(value)
       ,  index_(index)
       {
       }

    };

    typedef std::map<std::string, ValueAndIndex> SUPERSEEDS;

    std::auto_ptr<SUPERSEEDS>     superseeds_;

    std::string sourceFile_;
    int         sourceLine_;


    DECLARE_CXXTLS_EXCEPTION(TooLongException);
    DECLARE_CXXTLS_EXCEPTION(InvalidWordSyntaxException);
    DECLARE_CXXTLS_EXCEPTION(EmptyCommandException);

    BrashCommandLine(ARGUMENTS_LIST      *args,
                     REDIRECTION_LIST    *redirections,
                     SUPERSEEDS          *superseeds,
                     std::string const   &sourceFile,
                     int                  sourceLine
                    )
    :  arguments_(args)
    ,  redirections_(redirections)
    ,  superseeds_(superseeds)
    ,  sourceFile_(sourceFile)
    ,  sourceLine_(sourceLine)
         ///  Take ownership of the pointer objects passed in and delete them on destruction.
         ///  expression argument 0 is the command name
    {
    #ifdef DebugBuild
        if (args)
        {
            std::string arg0;

            (*args->begin())->isSimple(arg0);

        }
    #endif
    }

    BrashCommandLine(std::string::const_iterator &cmdLineStart, std::string::const_iterator const &cmdLineEnd);
        /// mainly for debug purposes, this constructor parses the command line
        /// and constructs all the needed parts.  It leaves cmdLineStart point to either cmdLineEnd or some
        /// token that should not be included in the command line:
        /// @li ;
        /// @li \n
        /// @li ;;
        /// @li &&
        /// @li |
        /// @li ||
        /// @li etc

    ~BrashCommandLine()
    {
        if(arguments_.get())
        {
            CXXTLS_FOREACH(cxxtls::StringExpression*  cur, *arguments_)
                delete cur;
        }

        if(redirections_.get())
        {
            CXXTLS_FOREACH(REDIRECTION_LIST::value_type &cur, *redirections_)
                delete cur.second;
        }

        if(superseeds_.get())
        {
             CXXTLS_FOREACH(SUPERSEEDS::value_type &cur, *superseeds_)
             {
                delete cur.second.value_;
                delete cur.second.index_;
             }
        }
    }

    cxxtls::StringExpression const* findInlineInputRedirection()
    {
          if(redirections_.get())
          {
              CXXTLS_FOREACH( REDIRECTION_LIST::value_type const &cur, *redirections_)
              {
                  std::string value;

                  if(cur.second && cur.second->isRedir() && cur.second->isSimple(value))
                  {

                     if(value.size() > 2)
                     {
                         
                         if(value[0] == '<' && value[1] == '<')
                            return cur.second;
                     }
                  }
              }
          }

          return 0;
    }


};

std::ostream &operator<< (std::ostream &s, BrashCommandLine const &me);  // standard output statement

#endif
