#include <string.h>

namespace cxxtls
{
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    struct Charset
        /// A character set object that represents the presence or
        /// absence of up to 256 character values.  
        ///
        /// Usage:
        ///
        ///    Charset   set;
        ///
        ///    std::string definition("a-z");
        ///
        ///    set.parse(definition);   
        ///  
        ///    if(set.test('a'))
        ///      cout << "good, it worked right" << endl;
        ///
        ///    if(set.test('1'))
        ///      cout << "bad! '1' should not be in the set" << endl;
        ///
    {
        char data_[32];  // 256 bits

        Charset()
        {
           memset(&data_[0], 0, sizeof(data_));
        }

        void setAll()
        {
           memset(&data_[0], 0xff, sizeof(data_));
        }

        void set(unsigned char bit, bool value)
            /// set the 'presence' bit for a given character value 
        {
           unsigned char byte = bit >> 3;   // divide the char value by 8

           unsigned char mask = ((unsigned char)(0x80)) >> (bit%8);

           if(value)
               data_[byte] |= mask;
           else
               data_[byte] &= ~(mask);

        }

        bool test(unsigned char bit) const
            /// test a specified character value to see if it appears
            /// in the character set.
        {
           unsigned char byte = bit >> 3;   // divide the char value by 8

           unsigned char mask = ((unsigned char)(0x80)) >> (bit%8);

           return    (data_[byte] & mask );
           
        }

        static unsigned char mapEscapeChar(unsigned char c)
            /// given a character which is meant to be the character following
            /// the '\\' in an escape sequence, return the interpreted value.
            ///
            /// Given the sequence "\t", pass the 't' to mapEscapeChar(), and
            /// back the tab character.
        {
             switch(c)
             {
                 case 's':  c = ' ';   break;
                 case 't':  c = '\t';  break;
                 case 'e':  c = 0x01b; break;
                 case 'a':  c = 0x07;  break;
                 case 'b':  c = 0x08;  break;
                 case 'r':  c = '\r';  break;
                 case 'n':  c = '\n';  break;

                 default:
                      break;
             }

             return c;
        }

        template<class String>
        void parse(String &s)
            ///  Parse a string as a character set.
        {
            typename String::iterator f = s.begin(),
                                      l = s.end();

            parse(f,l);
        }

        template<class Iterator, class constIterator>
        void parse(Iterator &f, constIterator const &l)
            ///  Parse character iterator range as a character set.
        {
            // parse a string of one of these forms:
            //
            //    "abcd"               -- the set containing each of the characters, a - d.
            //
            //    "a-zA-Z0-9_"         -- all letters and numbers and the underscore character.
            //
            //    "-+"                 -- "-" and "+"
            //
            //    "^a-z"               -- all characters but a-z!

            bool invertSet = false;

            if(f != l && *f == '^')
            {
                ++f;
                invertSet = true;
            }

            unsigned int prev='`';

            if(f != l && *f == '-')
            {
               prev='-';
               set(prev,true);
               ++f;
            }

            while(f != l)
            {
                unsigned char c = *f;

                if(c == '\\')
                {
                   ++f;

                   if(f == l)
                      set(c, true);
                   else
                   if(*f == '0' ) 
                   {
                      char digit;
                      unsigned char value=0;

                      while( f != l )
                      {
                          digit = *f;

                          if(digit >= '0' && digit <= '7')
                          {
                             value <<= 3;
                             value += digit - '0';
                          }
                          else
                              break;

                          ++f;
                      }

                      prev= value;
                      set(prev,true);
                      continue;             // f has alreadry been incremented


                   }
                   else
                   {
                      prev = mapEscapeChar(*f);
		      set(prev,true);
                      ++f;
		      continue;  // f is already incremented
                   }
                }
                if(c == '-')
                {
                   ++f;

                   if(f == l)
                   {
                       prev='-';
                       set(prev,true);
                   }
                   else
                   {
                      c = *f;

                      if(c == '\\')
                      {
                         ++f;

                         if(f !=l)
                         {
                             if(*f == '0')
                             {
                                 unsigned char value = 0;
                                 char digit;

                                 while(f != l)
                                 {
                                     // we'll go through here at least once

                                     digit = *f;

                                     if(   (digit >= '0')
                                        && (digit <= '7')
                                       )
                                     {
                                        value <<= 3;
                                        value += digit - '0';

                                        ++f;
                                     }
                                     else
                                        break;

                                 }

                                 --f;  // f will be ncremented later
				 c = value;
                             }
                             else
                                c = mapEscapeChar(*f);
                         }
                         else
                            --f;  // \\ was the last character
                      }

                      ++prev;

                      while(prev <= c)
                      {
                         set(prev,true);
                         ++prev;
                      }
                   }

                }
                else
                {
                    prev = c;
                    set(prev,true);
                }

                ++f; 


            }

            if(invertSet)
            {
                unsigned char *p = reinterpret_cast<unsigned char*>(&data_[0]);

                unsigned char *e = p + sizeof(data_);

                while(p != e)
                {
                   unsigned char v = ~(*p);

                   *p++ = v;

                }

            }

        }  // parse



    };
}
