#ifndef hash_list_h_included
#define hash_list_h_included


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.




#include <list>
#include <vector>
#include <utility>
#include <cxxtls/hashers.h>

namespace cxxtls
{

///@file
///
///  This file defines the hash_list container.


template<class Key, class Value, class Hash=Hasher<Key> >
class hash_list
//
///   A hash_list<Key,Value,Hash> is a list of pair<Key,Value> which can be quickly searched
///   for a given K. Most operations are of fixed cost.  Memory consumed is approximately 3 * sizeof(void*)
///   + sizeof(K) + sizeof(V).  The primary storage mechanism is a list of Key,Value,bucket_next
///   objects.  This gives O(1) insert, erase, iteration, etc.  The 'bucket_next' member kept
///   with the Key,Value pair lets us have a fixed time cost to find a given K -- there is a
///   hash table containing all Keys -- this table's buckets thread through the individual
///   objects stored in the linked list.
/// 
///   There can be only element in the container with a given Key value.
/// 
///   The template parameters are as follows:
/// 
///   @li  Key    -- an object which is the search key into the list.
///   @li  Value  -- an object which is assumed to be associated with the key.
///   @li  Hash -- a functor object with a function call operator method which
///               will give an 'int' given a 'Key'.  The performance of the
///               container is directly proportional to the uniqueness of the
///               mapping between keys and values.  If lots of Keys give the
///               same 'int' value, then performance of this class will be poor.
///               You probably want to user the template found in \<Hashs\>
///               for built in types, and follow the implementation examples
///               their for your own classes.
/// 
///   Here is an example use:
/// 
///   @code
///
///     struct IntHash  // or use Hasher<int> from file <hashers>
///     {
///        int operator() (int value ) { return value; }
///     };
/// 
///     typedef hash_list<int,string,IntHash> Container;
/// 
///     Container hl;
/// 
///     hl.insert(Container::mapped_data(9, "stuff"));
/// 
///     Container::iterator nine = hl.find(9);
/// 
///     string stuff = nine->second;
///
///   @endcode
///
///   Be careful when writing Hash's.  It is tempting to write a generic
///   Hash algorithm that simply takes any kind of object and adds up
///   the bytes in it.  This won't work because of alignment holes put into
///   the structure by the compiler -- and for other reasons.  You should make
///   the Hash's return value be dependent ONLY on the actual usuable fields
///   in a structure -- and it should be written to run very quickly.  A Hash's
///   function call operator can not be a static method If you have data in the
///   Hash, remember to make its copy constructor and destructor work correctly.
/// 
///   Here is a trivial example of hashing a string:
///
///   @code
//// 
///    struct StringHash  // or use Hasher<std::string> from file <hashers>
///    {
///      int operator() (string s)
///      {
///        int rv = 0;
/// 
///        string::const_iterator f = s.begin();
///        string::const_iterator l = s.end();
/// 
///        while(f != l) rv += *f++;
/// 
///        return ( (rv << 8) | s.size() );
/// 
///      }
///    };
///  @endcode
{
  Value empty_value_;  ///<  returned by the const array index operator when
		       ///  no object with a given key is specified

public:

  typedef hash_list<Key,Value,Hash>   self;

  typedef typename std::pair<Key,Value> mapped_type;
    ///<  
    ///  Data you should think about as being in the hash_list.
    ///  Actually, this is only part of the real 'value_type'

  struct value_type
    ///
    /// The objects actually stored in a hash_list are 'value_type'
    /// objects.  The member, data_, contains the pair<Key,Value> that
    /// is the user data provided by an insert operation.
    //
  {
    value_type *bucket_next_;  ///<  pointer to next value_type object in the hash bucket

    mapped_type data_;         ///<  user data

    value_type(value_type* bn, Key const &k, Value const &v)
    : bucket_next_(bn),
      data_(k,v)
    {
      /// construct a value_type given a key, a value, and a pointer to the
      /// next value_type object in the hash buck that this one goes in.
      ///
      /// @param bn first entry in the hash buck computed from this key (the key resolves to a bucket, and bn is the first entry currently there).
      /// @param k  the hash key used in the computations
      /// @param v  the value stored in the new hash bucket entry
    }

  };

  typedef std::list<value_type>    List;    ///<  the primary storage representationt ype
  typedef std::vector<value_type*> Vector;  ///< the collection of buckets

  typedef typename List::iterator        iterator;        ///<  std STL style iterator
  typedef typename List::const_iterator  const_iterator;  ///< std STL style const_iterator

private:

  size_t size_;                      ///<  cache length of list_
  size_t average_chain_length_;      ///<  when size_/ht_.size() \> acl, resize vector_
  List   list_;                      ///<  primary storage
  Vector ht_;                        ///<  hash buckets
  Hash hash;                         ///<  function which gives you hashes of the Key

public:

  hash_list(int chain_length=10)
  :  size_(0),
     average_chain_length_(chain_length),
     ht_(10)

     /// default constructor selects an resize chain length of 10 and
     /// 10 as the number of buckets in the hash table.
  {
     typename Vector::iterator first = ht_.begin();
     typename Vector::iterator last = ht_.end();

     while(first != last)
     {
	*first = 0;
	++first;
     }

  }

  ~hash_list()
    /// destructor
  {
  }

  size_t size() const { return size_; }  ///<  number of items in the container

  iterator begin() { return list_.begin(); } ///<  iterator to first element in the list
  iterator end()   { return list_.end();   } ///<  iterator to the end element of the list

  const_iterator begin() const { return list_.begin(); } ///<  iterator to first element in the list
  const_iterator end()   const { return list_.end();   } ///< iterator to end element of the list

  iterator insert(mapped_type const &data)
  ///
  /// Insert a Key/Value pair.  When inserting, you must supply a
  /// hash_list::mapped_type object.  Used like this:
  ///
  /// @code
  ///    hl->insert( hash_list::mapped_type(key,value) );
  /// @endcode
  ///
  //
  {
       list_.push_back( value_type(0, data.first, data.second) );

       iterator last = end();

       --last;


       value_type *p = & (*last);

       int hash_code = hash(data.first);

       hash_code %= ht_.size();

       last->bucket_next_ = ht_[hash_code];

       ht_[hash_code] = p;

       ++size_;

       if(size_ / ht_.size() > average_chain_length_)
       {
	 resize();
       }

       return last;
  }

  void erase(iterator location)
   
    /// Remove one element from the list at the specified location.
    /// Used like this:
    /// @code
    ///   hash_list<t>::iterator location = hl.find(some_key);
    ///
    ///   hl.erase(location);
    /// @endcode
  {
    value_type *p = &(*location);

    int bucket =  hash(p->data_.first) % ht_.size();

    if(ht_[bucket] == p)
    {
       // erasing first item in the bucket

       ht_[bucket] = p->bucket_next_;

    }
    else
    {
       // unlink the guy from inside the bucket's chain somewhere

       value_type *scan = ht_[bucket];

       while(scan->bucket_next_)
       {
	 if(scan->bucket_next_ == p)
	 {
	   scan->bucket_next_ = p->bucket_next_;
	   break;
	 }

	 scan = scan->bucket_next_;

       }

    }

    list_.erase(location);

    --size_;

  }

  iterator find(Key const &key)
    
    /// given a Key value, get an interator to the hash_list::value_type
    /// object that contains that key.
    
  {
    int hash_code = hash(key);

    hash_code %= ht_.size();

    value_type* scan = ht_[hash_code];

    while(scan)
    {
       if(scan->data_.first == key)
       {

	  // this code assumes that the value_type object is
	  // encased in a list_node which looks like this
	  //
	  //   list_node* next_;
	  //   list_node* prev_;
	  //   value_type v_;
	  //
	  // And that you can thus subtract 2 * sizeof list_node*
	  // from the address of the value type to get the address
	  // of the node type.
	  //
	  // It further assumes that an iterator is just a pointer
	  // to a list_node.
	  //
	  // There will have to be #if's for this logic on
	  // each platform.

	  iterator rv;

	  char *p = (char*)scan;

	  p -= 2 * sizeof(void*);

	  char **x = (char **)&rv;

	  x[0] = p;

	  return rv;
       }

       scan = scan->bucket_next_;
    }

    return end();

  }

  const_iterator find( Key const &index ) const
      /// @return an iterator to the entry with the specified key
      /// @param index the entry being searched for
  {
     self * me = (self*)(this);  // unconstify

     return me->find(index);

  }


  Value const &operator[] (Key const &index) const
    
    /// Array index operator.  returns a Value given its Key.
    /// When the hash_list is const, then this function just
    /// returns a const& to either the object in the list or
    /// to the empty_value_ object for this hash_list.
    ///
    /// @return a reference to the desired object
    /// @param index the key value of the desired object
    //
  {
    const_iterator rv = find(index);

    if(rv == end())
      return empty_value_;

    return rv->data_.second;
  }

  struct index_helper
    //
    /// When using a non-const hash_list, and invoking the
    /// array index operator, [], you get back an index_helper
    /// object rather than a reference to the Value object you
    /// would expect to be gettting.   This is because the behaviour
    /// of the array index operation requires that if the object
    /// isn't there already that it be inserted.  You could insert
    /// a 'default' Value object into the container, only to have it
    /// deleted if you are writing code like this:
    ///
    /// @li   list[stuff] = newValue;
    ///
    /// That is, if you are inserting into the list using the array
    /// index operator, you don't want to have to insert a default value,
    /// then remove it, then insert the newValue.  Instead, you want this
    /// operation to do just a single insert.
    ///
    /// The index_helper class aids in this goal.  Instead of having the
    /// non-const operator[] automatically insert 'default' Value object
    /// it returns immediately an index_helper object.  The index_helper
    /// object can be converted into Value const & and it has an operator=
    /// to handle assignments.
    ///
    /// For example, if you write code like this:
    ///
    /// @li list[stuff] = newValue
    ///
    /// You are getting the following:
    ///
    /// @li hash_list::operator[] returns index_helper(stuff,this);
    ///     index_helper::operator= is invoked with 'newValue' as a
    ///     parameter.
    ///
    /// On the other hand if you do this:
    ///
    /// @li Value v = list[stuff];
    ///
    /// hash_list::operator[] returns index_helper(stuff,this)
    /// and then v is assigned the value of index_helper::operator Value().
    ///
    /// Unfortunately this approach makes it hard if Value is a pointer
    /// because the following code won't really work:
    ///
    /// @li list[stuff]->member = 9;
    ///
    /// Instead you must write:
    ///
    /// @code
    ///  Value v = list[stuff];
    ///  v->member = 9;
    /// @endcoce
    ///
    /// Or
    /// @code
    ///    ((Value)(list[stuff]))->member = 9;  // YUCK!
    /// @endcode
    ///
  {
     typedef hash_list<Key, Value, Hash> HL;      ///< short name for this kind of hash list
     Key const&                            key_;  ///< Key of the object being referenced
     HL*                                   list_; ///< the hash_list being dereferenced

     index_helper(Key const& k, HL *l)
     : key_(k),
       list_(l)
     {
	///  plain vanila constructor
     }

     operator Value const &()
	//
	///  When doing this:
	///  @code 
	///    Value t = hash_list[key];
	///  @endcode
	///  You end up calling index_helper::operator Value const&()
	/// 
	/// 
     {
	iterator ptr = list_->find(key_);
	
	if(ptr == list_->end())
	  return list_->empty_value_;

	return ptr->data_.second;
	
     }


     Value const &operator=( Value const &r )
	//
	///   When you do this:
	///   @code
	///     hash_list[key] = newValue;
	///   @endcode
	///   Any old entry with this key/value pair is
	///   deleted and a new one is inserted
     {

	iterator ptr = list_->find(key_);
	
	if(ptr != list_->end())
	{
	   list_->erase(ptr);
	}
	
	list_->insert( mapped_type(key_,r) );
	
	return r;
	

     }


  };

  friend class index_helper;


  index_helper operator [] (Key const &index)
    ///
    /// Array index operator.  returns a Value given its Key.
  {
    // see the index_helper struct above for an explaination of
    // how this works

    return index_helper(index,this);
  }



private:

  void resize()
  {
    /// Increases the size of the hash table to speed up find operations.
    /// This recreates the mapping between items in the list and hash
    /// buckets to preserver the hashtable logic.

    int new_size = size_ / average_chain_length_;

    if(new_size < 10)
      new_size = 10;

    Vector new_table( new_size );

    while(new_table.size() != (unsigned)(new_size))
    {
      new_table.push_back(0);
    }

    ht_.swap(new_table);

    // now the hash table is larger but empty and the
    // bucket_next_ links in list_ are all wrong --
    // rebuild them all from scratch

    typename List::iterator f = list_.begin();
    typename List::iterator l = list_.end();

    while(f != l)
    {
      value_type *cur = &(*f);

      int hash_code = hash(cur->data_.first);

      hash_code %= ht_.size();

      cur->bucket_next_ = ht_[hash_code];

      ht_[hash_code] = cur;

      ++f;
    }

    // done!
  }
};
} // namespace cxxtls
#endif
