#ifndef SPLINE
#define SPLINE
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.
//
// Further warning, the algorithms container here are c++ transmogriphications
// of code from a college textbook -- originally fortran.  See below.

///@file
///  Declare the spline class -- which implements a cubic spline approximation
///  over a range.


#include <map>
#include <set>
#include <functional>
#include <utility>
#include <vector>
#include <stddef.h>

namespace cxxtls
{

//
//  This file defines the spline template and some classes that
//  make it more useful.  See them near the bottom of the file.
//

template<class X, class Y, class D=double>
class spline
  ///  
  ///  A spline is an approximation to a single valued function of
  ///  one variable.  The function is approximated by first defining a specific 
  ///  set of sample values.  Other data points of the function are interpolated
  ///  given the defined set.  That is, you can approximate data points that 
  ///  don't really exist using one the interpolate_* methods or the
  ///  function call operator.  See below.
  ///
  ///  The following interpolators exist:
  ///
  ///  @li  step function
  ///  @li  linear function
  ///  @li  cubic spline  (from "Numerical Computing: An introduction" by Shampine and Allen)
  ///
  ///  You can select which one this is using the
  ///  constructor or using the set_interpolator() method.  You can also call
  ///  any desired interpolator directly without using set_interpolator().
  ///
  ///  A 'sample' is a pair<X,Y>.  For a given X, only one
  ///  Y can be used.  X means 'independent variable' and
  ///  Y means 'dependent variable'.  That is, X is the domain
  ///  of a two dimensional graph point.  Y is the range.
  ///
  ///  The template parameter, D, lets you specify the data type to use when
  ///  doing multiplications and divisions and repeated operations.  Normally,
  ///  your X and Y will be one of:  int, float, double, short,
  ///  etc.  Using double as the default high resolution numeric data type
  ///  works fine for them.  It isn't used much anyway.  However, if you
  ///  supply your own user defined classes for X or Y, you may
  ///  need to supply your own D class as well.  Of course, you may not
  ///  need to, but the feature is there if you need it.
  ///
  ///  Interpolation is the act of guessing at values of the Y for
  ///  values of the X for which no specific data points are provided.
  ///  The following interpolation schemes are provided:
  ///
  ///  @li Step function approach.  The values between explicitly defined
  ///      values of the X are assumed to be the same value as the
  ///      explitly defined variable to the left.  The exceptions being:
  ///      If no data points are assumed the interpolated variable is undefined.
  ///      If the requested X is before the first defined value, then
  ///      the interpolated Y variable will be the same as the first Y.
  ///
  ///  @li Linear function approach.  In this approach, interpolation requests
  ///      for X values between extant samples will be linearly
  ///      interpolated.  For variables of X higher than the highest
  ///      extant sample, the line connecting the last two samples will be
  ///      extended.  For X requests lower than the first X, the
  ///      line connecting the first two points will be extended.  When only one
  ///      real sample is defined, that value will be returned by the interpolator
  ///      regardless of the X specified.
  ///    
  ///  @li Cubic spline approach.  In this approach, cublic spline curve fitting
  ///      will be done beween extant samples.  Interpolation around the ends
  ///      of the extant samples will be handled similarly to the linear
  ///      approach.  Derived from the example in the book
  ///
  ///      @code
  ///        Numerical Computing by Shampine and Allen.  Copyright 1973.
  ///        ISBN 0-7216-8150-6.
  ///      @endcode
  ///
  ///  The function all operator is provided so that you can easily use the
  ///  spline to access dependent variable values:
  ///
  /// 
  ///  @code
  ///     spline<int,int> s;
  ///
  ///     s.add_sample(99, 13);
  ///     s.add_sample(101, 20);
  ///
  ///     X x = 100;
  ///     Y y = s(x);
  ///
  ///  @endcode
  ///  y is now somewhere between 13 and twenty, given the current interpolator.
  ///  
  ///  Only one sample with a given X value will be kept.  If you add a new
  ///  sample at an old location, the old one will be replaced.  
  ///
  ///  Samples are stored in order, sorted by X.
  ///
  ///  Use method 'domain()' to get a list of all Xs found in samples
  ///  in this function.
  ///
  ///  You can use the apply method to apply an operation to all the actual
  ///  sample data points.  This lets you change them if you like.  
  ///
  ///  Note that the primary sampled storage class allows multiple samples of 
  ///  the same X but the add_sample method won't.  The multimap class 
  ///  is used instead of map because multimap supports the upperbound method
  ///  which can be used to quick find a location with the multimap NEAR
  ///  a given value of the X.  This avoids costly linear searches
  ///  when interpolating
  ///
  ///  The 'include' method lets add multiple samples in one operation -- there
  ///  a couple of variations.
  ///
  ///  The operators, +=, -=, *=, and /= let you include combine two splines
  ///  by interpolating this spline and another spline over all datapoints of
  ///  interest to either spline, and producing a new spline.  The interpolation
  ///  involved in these operator functions is the 'current' interpolator
  ///  set using set_interpolator() or as a parameter to the constructor.
  ///
  ///  Comments about the spline members can be found in the bodies thereof which
  ///  begin immediately after the class definition.
{

public:

  typedef X Xt;  ///<  alias for template parm
  typedef Y Yt;  ///<  alias for template parm
  typedef D Dt;  ///<  alias for template parm

  struct info  /// info stored for each x value in the sample data set
  {
    X y_; ///< value of the function at the specified x location

    D s_; ///< second derivative the interpolation function at x,
	  ///< possibly uninitialized if you aren't using the cubic
	  ///< spline.

    info(X y, D s=0)
    : y_(y),
      s_(s)
    {
    }

    operator X() const { return y_; }
  };

   typedef std::multimap<X,info>           samples_t;   ///< Primary storage
   typedef typename samples_t::value_type  sample_t;    ///< Elements stored

   typedef typename samples_t::iterator        iterator;        ///< iterator
   typedef typename samples_t::const_iterator  const_iterator;  ///< const_iterator


   typedef Y (spline::* interpolator_t) (X const &) const;
     /// interpolator_t  lets you pass the interpolator function
     /// as a pointer.

   typedef spline<X,Y,D> self;

private:

   samples_t       samples_;         ///< Map of actual data.
   interpolator_t  interpolator_;    ///< Current interpolator. 
   bool            dirty_;           ///< cubic spline s's need recalculation
   

   // Make sure you add all variables to the constructors, the swap method,
   // and the destructor if any.

public:
  
   inline void swap(self &rhs); 

   inline spline();
   inline spline(interpolator_t a);
   inline spline(self const& rhs);
   inline self &operator= (self const& rhs);

   inline void add_sample(X const &x, Y const &y) ;
   inline void add_sample(sample_t const &p); 

   inline interpolator_t interpolator() const ;
   inline void set_interpolator(interpolator_t t); 

   template<class SampleIterator>
     inline
       spline(SampleIterator first, SampleIterator last, interpolator_t t);

   template<class OtherKind>
     inline
       spline(OtherKind const &rhs, interpolator_t terp);

   inline Y operator() (X location) const;


   template<class Operation>
     void apply(Operation const&);  

   inline const_iterator ceil(X const& x) const;

   const_iterator floor(X const &x) const;

   // You can use the current interpolator using operator(), or you can
   // one of the following interpolators directly.

   Y interpolate_step (X const&) const;  
   Y interpolate_line (X const&) const;  
   Y interpolate_cubic(X const&) const;  

   inline const_iterator begin() const;
   inline const_iterator end()   const;

   inline sample_t front() const;  
   inline sample_t back() const ;  

   size_t size() const;

   void  domain(std::set<X> *target) const; ///< get all interesting Xs

   template<class Iterator> void include(Iterator, Iterator);

   template<class SF> inline void include(SF const &rhs); 

   template<class Other, class Operation>
     void combine(self &output, Other const&rhs, Operation const &op);

   template<class T>
     inline self &operator+= (T const &rhs);

   template<class T>
     inline self &operator-= (T const &rhs);

   template<class T>
     inline self &operator*= (T const &rhs);

   template<class T>
     inline self &operator/= (T const &rhs);

   template<class OutputIterator>
     void 
      evaluate(X start_x, 
	       X delta_x, 
	       OutputIterator begin,
	       OutputIterator end
	      ) const;

   void compute_coefficients();

};


/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
  inline 
  typename spline<X,Y,D>::const_iterator
  spline<X,Y,D>::
    begin() const  
  { 
    return samples_.begin(); 
  }
  
template<class X, class Y, class D>
  inline 
  typename spline<X,Y,D>::const_iterator
  spline<X,Y,D>::
    end() const  
  { 
    return samples_.end(); 
  } 

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
inline 
spline<X,Y,D>::
spline()
: interpolator_(&self::interpolate_step),
  dirty_(true)
{
   ///  Default constructor gives you the default interpolator (stepwise).
   ///  Creates an empty spline. Use add_sample to fill it.  You can of course
   ///  use any interpolator by calling it directly, or you can use 
   ///  set_interpolator to select a different one as the current interpolator.

}

/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
inline 
spline<X,Y,D>::
spline(interpolator_t a)
: interpolator_(a),
  dirty_(true)
{
  ///  Constructor that lets you specify the interpolator algorithm.
}

/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
inline 
spline<X,Y,D>::
spline(self const& rhs)
: interpolator_(rhs.interpolator_)
{
  /// Copy constructor.
  include(rhs);
}


/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
 template<class SampleIterator>
   inline
   spline<X,Y,D>::
   spline(SampleIterator first, 
	  SampleIterator last, 
	  typename spline<X,Y,D>::interpolator_t t
	 )
   : interpolator_(t),
     dirty_(true)
   {
     ///  Construct a spline from a pair of iterators to objects that look 
     ///  like samples.  That is, a range of pair<someXtype, someYtype>.
     ///  You must also specify the desired interpolator.

     include(first,last);

   }


/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
 template<class OtherKind>
   inline
   spline<X,Y,D>::
   spline(OtherKind const &rhs, 
	  typename spline<X,Y,D>::interpolator_t terp
	 )
   : interpolator_(terp),
     dirty_(true)
   {
      ///  Copy an extant interpolator's data.  You can specify a new
      ///  interpolator.

      include(rhs.begin(), rhs.end());
   }
/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
inline 
void 
spline<X,Y,D>::
swap(self &rhs)
   ///  Quickly swap the contents of this spline with that of another.
   ///  This occaisionally useful when writing algorithms that construct
   ///  objects and then destruct them.  You can arrange to have an empty
   ///  object destructed quickly while your constructed object gets
   ///  filled without a copy.  This of course requires thought.
{
  samples_.swap(rhs.samples_);

  std::swap    (dirty_,        rhs.dirty_);
  std::swap    (interpolator_, rhs.interpolator_);

}

/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
inline 
typename spline<X,Y,D>::self&
spline<X,Y,D>::
operator=( typename spline<X,Y,D>::self const& rhs)
   {
     ///  Assignment operator.

     samples_.erase(samples_.begin(), samples_.end());
     include(rhs);
     interpolator_ = rhs.interpolator_;
     return *this;
   }


/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
  inline 
  typename spline<X,Y,D>::const_iterator 
  spline<X,Y,D>::
  ceil(typename spline<X,Y,D>::Xt const& x) const 
     /// get an iterator to the sample whose X is immediately above 'x'.
     /// Returns end() if no such sample exists.
   { 
     return samples_.upper_bound(x); 
   } 

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
  inline 
  typename spline<X,Y,D>::Yt 
  spline<X,Y,D>::
    operator() (typename spline<X,Y,D>::Xt location) const 
   { 
     ///  Evaluates the function at X coordinate, location, using the
     ///  current interpolator.  You control the interpolator using
     ///  set_interpolator.  Or you can directly call the specific interpolator you
     ///  want.

     return (this->*interpolator_)(location); 
   }

/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
inline
void 
spline<X,Y,D>::
add_sample(X const &x, Y const &y) 
  /// Add a new sampled data point.  If there is an existing sample with
  /// the same X value, the Y for that sample will be replaced.  This
  /// overload lets you specify the without first constructing a sample_t.
{ 
     add_sample(sample_t(x,y)); 
}
/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
inline
void 
spline<X, Y,D>::
add_sample(sample_t const &p) 
  /// Add a new sampled data point.  If there is an existing sample with
  /// the same X value, the Y for that sample will be replaced.
{ 
   dirty_ = true;
   
   iterator where = samples_.find(p.first);

   if(where == samples_.end())
     samples_.insert(p); 
   else
     where->second = p.second;

} 

/////////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
   template<class Other, class Operation>
   void
   spline<X,Y,D>::
   combine(spline<X,Y,D> &output, Other const&rhs, Operation const &op)
     ///  Produce output by combining this with rhs according to op on a
     ///  point by point basis.  All sampled data points in both splines are
     ///  considered.  The operation is applied to each pair of points existing
     ///  in either spline.  Points missing in one are interpolated so as to
     ///  be able to combine them with the other to produce the output.
   {
     std::set<X> xs;

     domain(&xs);
     rhs.domain(&xs);

     self tmp(this->interpolator_);

     typename std::set<X>::const_iterator first = xs.begin(), last = xs.end();

     while(first != last)
     {
       X const &x = *first++;

       Y y = op( (*this)(x) , rhs(x) );

       tmp.add_sample(sample_t(x, y) );

     }

     tmp.swap(output);

   }
/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
  template<class Iterator>
   void 
   spline<X, Y, D>::
     include(Iterator l, Iterator g)
     //
     ///  Ensures that all samples in iterator range also appear in this 
     ///  sampled function.  Samples in the range l-g replace samples already
     ///  spline.
     ///
     ///  The iterators must point to pairs of some kind but those types
     ///  do not have to be exactly X and Y.  On the other hand, it must be
     ///  possible to convert them to X and Y respectively.
   {

     while(l != g) 
     { 
       sample_t tmp(l->first, static_cast<Y>(l->second) ); 
	   // if this line fails to compile
	   // it is because rhs is not
	   // compatible with this due to
	   // template parameter type issues

       add_sample(tmp);
       ++l;
     }
   }

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
   template<class SF>
     inline
     void
     spline<X, Y, D>::
      include(SF const &rhs)
     ///  Ensures that all samples in rhs also appear in this sampled function.
     ///  Samples in rhs superceed samples already in the spline.  rhs can be
     ///  of a different type of data structure but the structure's iterator
     ///  must point to pairs whose First and Second types are compatible with
     ///  X and Y.
   {
     include(rhs.begin(), rhs.end());
   }

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
inline
typename spline<X, Y, D>::interpolator_t 
spline<X, Y, D>::
interpolator() const 
  /// gets the current interpolator
{ 
  return interpolator_;  
}

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
inline
void 
spline<X,Y,D>::
set_interpolator(interpolator_t t)
{ 
  interpolator_ = t; ///< Change to a new interpolator
  dirty_ = true;
}

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
inline
typename spline<X, Y, D>::sample_t
spline<X,Y,D>::
front() const
  ///  Returns the first sample or an undefined value if there are none.
{
  if(begin() != end())
    return *begin();
  
  return sample_t(X(), Y());
}

template<class X, class Y, class D>
inline
typename spline<X, Y, D>::sample_t
spline<X,Y,D>::
back() const
  ///  Returns the last sample or an undefined value if there are none.
{
  if(begin() != end())
  {
    const_iterator e = end();
    --e;

    return *e;
  }
  
  
  return sample_t(X(), Y());
}


template<class X, class Y, class D>
typename spline<X, Y, D>::const_iterator
spline<X,Y,D>::
floor(X const& x) const
  //
  ///  Returns the iterator to the sample whose X component is
  ///  as high as possible but <= x.  Returns end() if no such sample
  ///  exists.  Consider the sequence:
  ///  @code
  ///    10, 11, 12
  ///  @endcode
  ///  If you ask for 9, you will get end() returned.  If you ask
  ///  for 13, you will get an iterator to 12.  If you ask for 11.5,
  ///  you will get an iterator to 11.

{
  const_iterator ub    = samples_.upper_bound(x);
  const_iterator begin = samples_.begin();

  while(ub != begin)
  {
    --ub;

    if(ub->first <= x)
      return ub;

  }

  return samples_.end();

}

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
Y
spline<X, Y,D>::
interpolate_step(X const& x) const
  ///  This method interpolates the value of the sampled function at 'x',
  ///  returning the corresponding y.  See comments in it for a description
  ///  of how the interpolation works, but generally, this function treats
  ///  the sampled function as a step wise constant function that looks like
  ///  this:
  /// @code
  ///
  /// D|                         +------------->
  ///  |                         |
  /// C|         +-----------+   |
  /// B| <-+-----+           |   |
  /// A|                     +---+
  ///  +-------------------------------------------
  ///      x0    x1          x2  x3  
  ///                 u                  v
  /// @endcode
  ///
  ///  In this diagram, the samples are (x0,B), (x1,C), (x2,A), (x3,D).
  ///  A step wise interpolator makes sense when tracking inventory and the
  ///  time to load and unload trucks is insignificant.  It is the fastest
  ///  interpolator. 
  ///
  ///  The return value is calculated like this:  If there are no samples,
  ///  the return value is undefined.  If the requested 'x' value to be
  ///  interpolated is less than the lowest sampled x value, x0, then you
  ///  get x0's value functional value (B).  Otherwise you get the value
  ///  of the actual sample data, Y component, of the sample on or to
  ///  the immediate left.  So at point u, you would get C, and at v you would
  ///  get D.
  ///
  ///  If there are no samples at all, you get an undefined result.
  ///  
{
  const_iterator loc = floor(x);

  if(loc == end())
  {
    if(begin() == end())
      return Y();           // so samples at all, return empty data

    return begin()->second;
  }

  return loc->second;

}

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
Y
spline<X, Y,D>::
interpolate_line(X const& x) const
  //
  ///  This method interpolates the value of the sampled function at 'x',
  ///  returning the corresponding y.  See comments in it for a description
  ///  of how the interpolation works, but generally, this function treats
  ///  the sampled function as a piecewise linear function, like this:
  ///  @code
  ///
  /// D|         +                                                         ;
  ///  |        / \                                                        ;
  /// C|       /   \             +-->                                      ;
  ///  |      /     \           /                                          ;
  ///  |     /       \         /                                           ;
  ///  |    /         \       /                                            ;
  /// B| <-+           \     /                                             ;
  /// A|                +---+                                              ;
  ///  +-------------------------------------------                        ;
  ///      x0    x1     x2  x3   x4                                        ;
  ///               u                  v                                   ;
  ///
  ///  @endcode
  ///  In this diagram, the samples are (x0,B), (x1,D), (x2,A), (x3,A), (x4,C).
  ///  This is the second fastest interpolator and is usuable to represent the
  ///  manufacturing of objects.  Usually, an object takes a certain 
  ///  manufacturing time -- the more time, the more objects are made.  Sales
  ///  estimates generally work like this.  Slowly changing real world physical
  ///  phenomena can also be modeled this way.  If you need more accuracy,
  ///  use more samples.
  ///
  ///  The return value is calculated like this:  If there are no samples,
  ///  the return value is undefined.  If the requested 'x' value to be
  ///  interpolated is less than the lowest sampled x value, x0, then you
  ///  get x0's value functional value (B).  If the requested x value is greater
  ///  than the highest sample's X value (x4), then you will get the
  ///  Y of the highest sample.  This approach allows you set your value
  ///  for the 'out of range' situations.  You just jam in extra samples to
  ///  represent the too low and too high situations.  Otherwise, if you
  ///  request a location between two extant samples, you will get a linearly
  ///  approximated value.  Consider point 'u' in the above picture.  u likes
  ///  between x1 and x2.  The Y associated with us is calculated like
  ///  this:
  ///  @code
  ///    slope = (B - D) / (x2 - x1)  // negative value of course
  ///
  ///    deltax = x(u) - x1
  ///
  ///    y(u) = B + slope * deltax
  ///  @endcode
  ///  If there are no samples at all, you get an undefined result.
  ///  
{
  const_iterator loc = floor(x);

  if(loc == end())
  {
    // either no samples or all higher than x

    if(loc == begin())
    {
      return Y();  // no data to use
    }

    return begin()->second;  // return lowest known value

  }

  const_iterator ub = loc;
  ++ub;

  if(ub == end())
  {
    // x is >= last sample -- return the last samples value

    return loc->second;
  }

  //
  //  The requested X value lies somewhere between the samples pointed to
  //  by loc and ub.
  //

  // capital names represent highres numbers

  D  OLD_Y  = loc->second;    // OLD_X and Y represent floor sample ;>
  D  OLD_X  = loc->first;
  D  X1     = D(x);
  D  DELTAX = D(ub->first) - OLD_X;

  if(DELTAX == 0)
   return static_cast<Y>(OLD_Y);

  D  DELTAY = D(ub->second) - OLD_Y;
  D  SLOPE  = DELTAY / DELTAX;

  return static_cast<Y>(OLD_Y + SLOPE * (X1 - OLD_X));

}

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
template<class Operation>
void 
spline<X,Y,D>::
apply(Operation const &op)
  //
  /// Modifies the extant samples by visiting each one and replacing the
  /// the Y in the sample with the return value from the operation.
  /// The operation must take a sample_t as a parameter. 
  //
{

   iterator first = samples_.begin(),
	    last  = samples_.end();

   while(first != last)
   {
     first->second = op(*first);
     ++first;
   }

}

/////////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
void
spline<X, Y,D>::
domain(std::set<X>* output) 
const
  //
  /// Get a set of the sampled data X components (ie the domain locations
  /// where samples are taken.
  //

{
  const_iterator l = samples_.begin(),
		 g = samples_.end();
  while(l != g )
  {
    output->insert(l->first);
    ++l;
  }
}

/////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
  template<class T>
  inline
  spline<X,Y,D>& 
  spline<X,Y,D>::
    operator+= (T const &rhs)
       ///  Add another spline to this one.  Make sure that all samples in both
       ///  splines are addressed.
     {
       self tmp;
    
       combine(tmp, rhs, std::plus<Y>()); 
    
       swap(tmp);
    
       return *this;
     }

////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
 template<class T>
   inline
   spline<X,Y,D>&
   spline<X,Y,D>::operator-= (T const &rhs)
     ///  Add another spline to this one.  Make sure that all samples in both
     ///  splines are addressed.
   {
     self tmp;

     combine(tmp, rhs, std::minus<Y>()); 

     swap(tmp);

     return *this;
   }

////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
 template<class T>
 inline 
 spline<X,Y,D>&
 spline<X,Y,D>::
   operator*= (T const &rhs)
     ///  Add another spline to this one.  Make sure that all samples in both
     ///  splines are addressed.
   {
     self tmp;

     combine(tmp, rhs, std::multiplies<Y>()); 

     swap(tmp);

     return *this;
   }

////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
  template<class T>
   inline
   spline<X,Y,D>&
   spline<X,Y,D>::
     operator/= (T const &rhs)
       ///  Add another spline to this one.  Make sure that all samples in both
       ///  splines are addressed.
     {
       self tmp;
     
       combine(tmp, rhs, std::divides<Y>()); 
     
       swap(tmp);
     
       return *this;
     }

////////////////////////////////////////////////////////////////////
template<class X, class Y, class D>
  template<class OutputIterator>
  void 
  spline<X,Y,D>::
  evaluate(X start_x, 
	   X delta_x, 
	   OutputIterator begin,
	   OutputIterator end
	  ) const
    //
    ///  Create a collection of interpolated values.  Output those
    ///  values through an iterator that points to a pair<X,Y> where 
    ///  X, and Y are assignment compatible with double.  The begin,end
    ///  variables define the standard output iterator range logic.  The
    ///  first value output will be (start_x, f(start_x)), the second will be
    ///  (start_x+delta_x, f(start_x+delta_x)), and so forth.
  {
     while(begin != end)
     {
       begin->first  = start_x;
       begin->second = (this->*interpolator_)(start_x);

       ++begin;

       start_x += delta_x;
     }
  }
////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
void
spline<X,Y,D>::
compute_coefficients()
  ///  Compute the second derivative of the interpolation functions
  ///  and store them in the s_ member of the info associated with every
  ///  x/y pair.
{
  using std::vector;

  dirty_ = false;

  size_t n = samples_.size();

  // copy the map to arrays, doh!

  vector<D> xn(n+1), fn(n+1), rho(n+1), tau(n+1), s(n+1); 

    {
      const_iterator first = samples_.begin(),
		     last  = samples_.end();

      typename vector<D>::iterator xout = xn.begin() + 1;
      typename vector<D>::iterator yout = fn.begin() + 1;

      while(first != last)
      {
	*xout++ = first->first;
	*yout++ = first->second.y_;

	++first;
      }

    }

  // the two vectors, xn and fn are have undefined values in location 0

  size_t nm1 = n-1;
  
  rho[2] = 0;
  tau[2] = 0;

  int i;

  for(i=2; i <= static_cast<int>(nm1); ++i)
  {
    size_t iim1 = i-1;
    size_t ii   = i;
    size_t iip1 = i+1;

    D him1 = xn[ii] - xn[iim1];
    D hi   = xn[iip1]-xn[ii];
    D temp = (him1/hi)*(rho[i]+2)+2;
		 		   
    rho[i+1]=-1.0/temp;

    D d = 6.0 *((fn[iip1]-fn[ii])/hi-(fn[ii]-fn[iim1])/him1)/hi;

    tau[i+1] = (d-him1*tau[i]/hi)/temp;

  }

  s[1] = 0;
  s[n] = 0;

  int nm2 = n-2;

  for(i=1; i <= nm2; ++i)
  {
    size_t ib = n-i;
    s[ib] = rho[ib+1] * s[ib+1] +tau[ib+1];
  }

  {
    typename vector<D>::iterator first = s.begin() + 1,
				 last  = s.end();

    iterator output = samples_.begin();

    while(first != last)
    {
      output->second.s_ = *first++;
      ++output;
    }

  }


}
////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
Y
spline<X,Y,D>::
interpolate_cubic(X const& x) const
{
  using std::vector;

  if(dirty_)
  {
    self* me = const_cast<self*>(this);
    me->compute_coefficients();
  }

  size_t n = samples_.size();
  
  if(n == 0)
    return 0;

  if(samples_.size() == 1)
  {
    return samples_.begin()->second.y_;
  }

  if( x < samples_.begin()->first )
  {
    const_iterator lb = samples_.begin();
    const_iterator ub = lb; ++ub;

    D m = (ub->second.y_ - lb->second.y_) / 
		        (ub->first     - lb->first );

    D deltax = lb->first  - x;
    D deltay = m * deltax;

    return lb->second.y_ - deltay;
  }

  const_iterator end = samples_.end(); --end;

  if( x >= end->first)
  {
    const_iterator ub = samples_.end(); --ub;
    const_iterator lb = ub; --lb;

    D m = (ub->second.y_ - lb->second.y_) / 
		        (ub->first     - lb->first );

    D deltax = x - ub->first ;
    D deltay = m * deltax;

    return ub->second.y_ + deltay;
  }

  const_iterator ilp1 = ceil(x);
  const_iterator il   = floor(x);

  D a = ilp1->first - x;
  D b = x - il->first;
  D h = ilp1->first - il->first;

  return  a * il->second.s_ * (a * a / h - h)/6.0 +
		   b * ilp1->second.s_ * (b * b / h - h)/6.0 +
		   ( a * il->second.y_ + b * ilp1->second.y_)/h;

}
////////////////////////////////////////////////////////////////////

template<class X, class Y, class D>
  inline
  size_t
  spline<X,Y,D>::
  size() const
  {
    return samples_.size();
  }

////////////////////////////////////////////////////////////////////

} // namespace cxxtls

#endif
